/* version.h - version defines for lmin */

#ifndef _VERSION_H_
#define _VERSION_H_

#define LMIN_PROGRAM "lmin"
#define LMIN_VERSION "1.00 02Sep2026"
#define LMIN_URL "http://www.acme.com/software/lmin/"

#endif /* _VERSION_H_ */
